from typing import List

from jellyfish import metaphone 
from numpy import ndarray, array, argmax

from ffast.poincare.utils import Poincare

class Token:
    def __init__(self, raw_token:str, normalised_token:str, vector:ndarray, id:int) -> None:
        self.text = raw_token
        self.morphology = normalised_token
        self.phonology = metaphone(raw_token)
        self.semantics = vector
        self.id = id

    def __repr__(self) -> str:
        return f"""
        text = {self.text}
        morphology = {self.morphology}
        phonology = {self.phonology}
        id = {self.id}
        """

    def __str__(self) -> str:
        return self.text 

    def most_similar(self, others:List["Token"]) -> "Token":
        others_vectors = array(list(map(lambda other:other.semantics, others)))
        index_best = argmax(others_vectors@self.semantics)
        return others[index_best]


from typing import List, Generator

from nltk.util import ngrams
from numpy import zeros, ndarray, argmax, full

from ffast.poincare.token import Token
from ffast.poincare.tokens import Tokens
from ffast.poincare.utils import Poincare, PREPROCESSOR, VOCABULARY, VECTORS, SIZE_SENTENCE_VECTOR

class Tokeniser:
    def __init__(self) -> None:
        self.size = SIZE_SENTENCE_VECTOR
        self.token_size = Poincare.SIZE_VECTOR.value
        self.special_tokens = list()

    def __len__(self) -> int:
        return Poincare.SIZE_VOCABULARY.value + len(self.special_tokens) + 1
    
    def add_special_token(self, token:str) -> None:
        if token not in self.special_tokens and token not in VOCABULARY:
            self.special_tokens.append(token)
            self.special_tokens.sort()

    def encode(self, text:str) -> Tokens:
        return Tokens(self._tokenise(text),pad_token_id=len(self)-1)
    
    def decode(self, ids:List[int]) -> Tokens:
        return Tokens(list(self._convert_ids_to_tokens(ids)),pad_token_id=len(self)-1)
    
    def _tokenise(self, text:str) -> List[Token]:
        unknown_token_id = len(self)-1
        words = text.split()
        number_of_words = len(words)
        tokens = [None]*number_of_words
        for ngram_size in range(number_of_words+1,0,-1):
            for index_start,ngram in enumerate(ngrams(words,ngram_size)):
                index_end = index_start + ngram_size
                if any(tokens[index_start:index_end]):
                    continue 
                raw_token = ' '.join(words[index_start:index_end])
                normalised_token = PREPROCESSOR.normalise(text=' '.join(ngram))
                if normalised_token in self.special_tokens:
                    id = Poincare.SIZE_VOCABULARY.value + self.special_tokens.index(normalised_token)
                    vector = full(Poincare.SIZE_VECTOR.value,-1)
                elif raw_token in VOCABULARY:
                    id = VOCABULARY.index(raw_token)
                    vector = VECTORS[id]
                elif normalised_token in VOCABULARY:
                    id = VOCABULARY.index(normalised_token)
                    vector = VECTORS[id]
                else:
                    id = unknown_token_id
                    vector = zeros(Poincare.SIZE_VECTOR.value)
                    normalised_token = Poincare.UNKNOWN.value
                if id != unknown_token_id or ngram_size==1:
                    tokens[index_start:index_end] = [Poincare.SKIP.value]*ngram_size
                    tokens[index_start] = Token(
                        raw_token=raw_token,
                        normalised_token=normalised_token,
                        vector = vector,
                        id = id
                    )
        return list(filter(lambda token:isinstance(token,Token),tokens))

    def _convert_ids_to_tokens(self, ids:List[int]) -> Generator[Token,None,None]:
        for id in ids:
            if id < Poincare.SIZE_VOCABULARY.value:
                token = VOCABULARY[id]
                vector = VECTORS[id]
            else:
                id_special_token = id - Poincare.SIZE_VOCABULARY.value
                if id_special_token < len(self.special_tokens):
                    token = self.special_tokens[id_special_token] 
                else:
                    token = Poincare.UNKNOWN.value
                vector= zeros(Poincare.SIZE_VECTOR.value)
            yield Token(
                raw_token=token,
                normalised_token=token,
                vector = vector,
                id = id
            )

    @staticmethod
    def decode_semantics(poincare_vectors:List[ndarray]) -> Tokens:
        return Tokeniser.decode(ids=map(Tokeniser._convert_semantics_to_token_id,poincare_vectors))

    @staticmethod
    def _convert_semantics_to_token_id(semantics:ndarray) -> int:
        return argmax(VECTORS @ semantics)


from string import punctuation

from nltk.stem import WordNetLemmatizer
from unidecode import unidecode

class PreprocessingPipeline:
    def __init__(self) -> None:
        self.lemmatiser = WordNetLemmatizer()
        self.TRANSLATION_TABLE = "".maketrans(punctuation, " " * len(punctuation))

    def preprocess(self, text:str) -> str:
        lowercase_text = text.lower()
        unaccented_text = unidecode(lowercase_text)
        unpunctuated_text = unaccented_text.translate(self.TRANSLATION_TABLE)
        return unpunctuated_text

    def normalise(self, text:str) -> str:
        text = self.preprocess(text)
        for delimiter in "-_":
            text = text.replace(delimiter," ")
        return ' '.join(map(self.lemmatiser.lemmatize,text.split()))



from enum import Enum
from pathlib import Path 

from numpy import loadtxt

from ffast.preprocessor import PreprocessingPipeline

PREPROCESSOR = PreprocessingPipeline()
METAPHONES = "ABCEFHIJKLMNOPRSTUWXY0. "
SIZE_METAPHONES = len(METAPHONES)
PATH = Path(__file__).parent/"poincare.txt"
raw_vocab = loadtxt(PATH,usecols=0,dtype=str)
VOCABULARY = list(map(PREPROCESSOR.normalise,raw_vocab))
VECTORS = loadtxt(PATH,usecols=range(1,101))

class Poincare(Enum):
    SIZE_VECTOR = 100
    SIZE_VOCABULARY = len(VOCABULARY)
    UNKNOWN = "<Unknown>"
    SKIP = "skip"

SIZE_TOKEN_VECTOR = Poincare.SIZE_VECTOR.value + SIZE_METAPHONES
SIZE_POWER_MEANS_VECTOR = 5*SIZE_TOKEN_VECTOR
SIZE_SENTENCE_VECTOR = 2*SIZE_POWER_MEANS_VECTOR


from typing import List
from math import sin,cos

from scipy.stats import gmean, hmean
from scipy.fft import fftn
from numpy.linalg import norm
from numpy import (
    ndarray, concatenate, argmax, array,
    zeros, mean, sum, pad
)
import numpy

from ffast.poincare.token import Token
from ffast.poincare.utils import Poincare, METAPHONES, SIZE_METAPHONES
from ffast.wordnet.utils import STOPWORDS

class Tokens:
    def __init__(self, tokens:List[Token], pad_token_id:int) -> None:
        self.pad_token_id = pad_token_id
        self.tokens = tokens
        self.__pointer = -1
        self.ids = list(map(lambda token:token.id,self))
        self.vector = self.__sentence_embedding() if any(tokens) else None

    def __repr__(self) -> str:
        return '\n'.join(map(repr,self.tokens))
    
    def __str__(self) -> str:
        return ' '.join(map(str,self.tokens))

    def __len__(self) -> int:
        return len(self.tokens)

    def __iter__(self) -> "Tokens":
        return self
    
    def __next__(self) -> Token: 
        self.__pointer += 1
        if self.__pointer < len(self):
            return self.tokens[self.__pointer]
        raise StopIteration
    
    def __getitem__(self,index:int) -> Token:
        return self.tokens[index]

    def __round__(self, n:int) -> ndarray:
        return pad(self.ids, pad_width=(0,n), constant_values=self.pad_token_id)[:n]

    def semantics(self) -> List[ndarray]:
        return list(map(lambda token:token.semantics, self.tokens))
    
    def weights(self) -> List[float]:
        return list(map(self.__weight,self.tokens))

    def skip_common_words(self) -> "Tokens":
        return Tokens(list(filter(lambda token:self.__weight(token)>0.,self.tokens)),pad_token_id=self.pad_token_id)

    def skip_unknowns(self) -> "Tokens":
        return Tokens(list(filter(lambda token:token.morphology != Poincare.UNKNOWN.value,self.tokens)),pad_token_id=self.pad_token_id)

    def most_similar(self, others:List["Tokens"]) -> "Tokens":
        others_vectors = array(list(map(lambda other:other.vector, others)))
        index_best = argmax(others_vectors@self.vector)
        return others[index_best]

    def __sentence_embedding(self) -> ndarray:
        sparse_token_vectors = list(map(self.__embed_token,self.tokens))
        position_vectors = list(map(self.__embed_position,range(len(self.tokens))))
        contextual_token_vectors = sum([sparse_token_vectors,position_vectors],axis=0)
        dynamics_of_token_vectors = self.__fourier_transformation(sparse_token_vectors)
        return concatenate([
            self.__power_means(contextual_token_vectors),
            self.__power_means(dynamics_of_token_vectors)
        ])
        
    @staticmethod
    def __power_means(vectors:List[ndarray]) -> ndarray:
        positive_vectors = list(map(abs,vectors))
        return concatenate([
            numpy.max(vectors,axis=0),
            numpy.min(vectors,axis=0),
            mean(vectors,axis=0),
            gmean(positive_vectors),
            hmean(positive_vectors),
        ])
    
    @staticmethod
    def __fourier_transformation(vectors:List[ndarray]) -> ndarray:
        return abs(fftn(vectors))

    @staticmethod 
    def __embed_position(position:int) -> ndarray:
        SIZE_VECTOR = SIZE_METAPHONES + Poincare.SIZE_VECTOR.value
        position_vector = zeros(SIZE_VECTOR)
        for index in range(0,SIZE_VECTOR-1,2):
            angle = position/(1e5 ** ((2*index)/SIZE_VECTOR))
            position_vector[index]=sin(angle)
            position_vector[index+1]=cos(angle)
        return position_vector

    @staticmethod
    def __embed_token_phonetics(token:Token) -> ndarray:
        embedding = zeros(SIZE_METAPHONES)
        for character in token.phonology:
            metaphone_index = METAPHONES.index(character)
            embedding[metaphone_index] = 1
        return embedding

    @staticmethod
    def __embed_token_semantics(token:Token) -> ndarray:
        return token.semantics

    @staticmethod
    def __embed_token(token:Token) -> ndarray:
        return concatenate([
            Tokens.__embed_token_phonetics(token),
            Tokens.__embed_token_semantics(token),
        ])    
    
    @staticmethod
    def __weight(token:Token,minimum_weight:float = 0.,maximum_weight:float = 1.,scaling_factor:float = 10.) -> float:
        if token.text in STOPWORDS:
            return 0.
        approximation_of_token_frequency = .5 - norm(token.semantics)
        approximation_of_token_frequency *= scaling_factor
        return min(maximum_weight,max(minimum_weight,approximation_of_token_frequency))


__version__ = "0.4.6"
WORDNET = "wordnet"
POINCARE = "poincare"

def load(vectors:str=WORDNET):
    if vectors==WORDNET:
        from ffast.wordnet.tokeniser import Tokeniser
        return Tokeniser()
    if vectors==POINCARE:
        from ffast.poincare.tokeniser import Tokeniser
        return Tokeniser()    
    raise TypeError(f"{vectors} is an unrecognised choice. Valid choices are: '{WORDNET}' or '{POINCARE}'")


from typing import List, Generator, Optional, Set

from ffast.wordnet.utils import (
    WordNet, VOCABULARY, STOPWORDS,
    SIZE_STOPWORDS, SIZE_WORDNET
)

from nltk.corpus.reader.wordnet import Synset
from nltk.wsd import lesk
from nltk.corpus import wordnet
from jellyfish import metaphone 

class Token:
    def __init__(
        self, raw_token:str, normalised_token:str, 
        synset_name:Optional[str],SPECIAL_TOKENS:Optional[List[str]],
        sense_disambiguation_context:Optional[str]=None
    ) -> None:
        self.text = raw_token 
        self.morphology = normalised_token
        self.phonology = metaphone(raw_token)
        self.similar_tokens = set()
        self.opposite_tokens = set()
        self.related_tokens = set()
        self.semantics = set()
        self.definition = None
        self.example = None
        if raw_token in SPECIAL_TOKENS:
            self.id = SIZE_WORDNET + SIZE_STOPWORDS + SPECIAL_TOKENS.index(raw_token)
            self.tag = WordNet.SPECIAL.value
        elif normalised_token in STOPWORDS: 
            self.id = SIZE_WORDNET + STOPWORDS.index(normalised_token)
            self.tag = WordNet.STOPWORD.value            
        elif synset_name is None:
            self.id = SIZE_WORDNET + SIZE_STOPWORDS + 1
            self.tag = WordNet.UNKNOWN.value 
        else:
            if sense_disambiguation_context is None:
                meaning = wordnet.synset(synset_name)
                self.id = VOCABULARY.index(synset_name)
            else:
                meaning = lesk(sense_disambiguation_context.split(),synset_name.split(WordNet.SYNSET_NAME_DELIMITER.value)[0])
                self.id = VOCABULARY.index(meaning.name())
            self.tag = WordNet.POS_MAP.value.get(meaning.pos())
            self.similar_tokens = self.synonyms(normalised_token,meaning)
            self.opposite_tokens = set(self.antonyms(meaning))
            self.related_tokens = self._related_to(meaning)
            self.semantics = self.synset_lineage(meaning)
            self.definition = meaning.definition()
            self.example = WordNet.EXAMPLE_DELIMITER.value.join(meaning.examples())

    def __repr__(self) -> str:
        return f"""
        text = {self.text}
        morphology = {self.morphology}
        phonology = {self.phonology}
        id = {self.id}
        tag = {self.tag}
        similar = {self.similar_tokens}
        opposite = {self.opposite_tokens}
        related = {self.related_tokens}
        semantics = {self.semantics}
        definition = {self.definition}
        example = {self.example}
        """

    def __str__(self) -> str:
        return self.text 
        
    def _related_to(self, meaning:Synset) -> Set[str]:
        return set(self._remove_duplicates(
            values=self._get_related_meanings(meaning),
            duplicates={self.morphology} | self.similar_tokens
        ))

    def _similarity(self, other:"Token") -> int:
        return len(
            self.semantics.intersection(other.semantics)
        ) + len(
            self.similar_tokens.intersection(other.similar_tokens)
        ) + len(
            self.related_tokens.intersection(other.related_tokens)
        ) + len(
            set(self.phonology).intersection(other.phonology)
        ) + int(self.id == other.id) + int(self.tag==other.tag)
    
    def most_similar(self, others:List["Token"]) -> "Token":
        scores = list(map(self._similarity, others))
        max_score = max(scores)
        index_best = scores.index(max_score)
        return others[index_best]

    @staticmethod
    def synset_lineage(meaning:Synset) -> Set[str]:
        return set(map(
            lambda parent:parent.name(), 
            meaning.closure(lambda parent:parent.hypernyms())
        ))

    @staticmethod
    def synonyms(token:str, meaning:Synset) -> Set[str]:
        return set(Token._remove_duplicates(
            values=meaning.lemma_names(),
            duplicates={token}
        ))

    @staticmethod
    def antonyms(meaning:Synset) -> Generator[str,None,None]:
        for name in meaning.lemmas():
            for antonym_meaning in name.antonyms():
                yield antonym_meaning.name()

    @staticmethod
    def _get_related_meanings(meaning:Synset) -> Generator[str,None,None]:
        for related_meanings in (
            meaning.hyponyms(),
            meaning.part_meronyms(),
            meaning.substance_meronyms(),
            meaning.member_meronyms(),
            meaning.part_holonyms(),
            meaning.substance_holonyms(),
            meaning.member_holonyms(),
            meaning.topic_domains(),
            meaning.region_domains(),
            meaning.usage_domains(),
            meaning.entailments(),
            meaning.causes(),
            meaning.also_sees(),
            meaning.verb_groups(),
            meaning.similar_tos()
        ):
            for related_meaning in related_meanings:
                for name in related_meaning.lemma_names():
                    yield name

    @staticmethod
    def _remove_duplicates(values:List[str],duplicates:Set[str]) -> Generator[str,None,None]:
        for value in set(values):
            if value not in duplicates:
                yield value


from typing import List, Generator, Optional
from functools import reduce 

from numpy import ndarray, concatenate, argmax, array, zeros, pad

from ffast.wordnet.token import Token
from ffast.wordnet.utils import (
    WordNet, METAPHONES, VOCABULARY,
    SIZE_SEMANTIC_VECTOR, SIZE_METAPHONES
)

class Tokens:
    def __init__(self, tokens:List[Token], pad_token_id:int) -> None:
        self.pad_token_id = pad_token_id
        self.tokens = tokens
        self.__pointer = -1
        self.ids = list(map(lambda token:token.id,self))
        self.vector = self.__sentence_embedding() if any(tokens) else None

    def __repr__(self) -> str:
        return '\n'.join(map(repr,self.tokens))
    
    def __str__(self) -> str:
        return ' '.join(map(str,self.tokens))

    def __len__(self) -> int:
        return len(self.tokens)

    def __iter__(self) -> "Tokens":
        return self
    
    def __next__(self) -> Token: 
        self.__pointer += 1
        if self.__pointer < len(self):
            return self.tokens[self.__pointer]
        raise StopIteration
    
    def __getitem__(self,index:int) -> Token:
        return self.tokens[index]

    def __round__(self, n:int) -> ndarray:
        return pad(self.ids, pad_width=(0,n), constant_values=self.pad_token_id)[:n]

    def skip_unknowns(self) -> "Tokens":
        return Tokens(list(filter(lambda token:token.tag != WordNet.UNKNOWN.value,self.tokens)),pad_token_id=self.pad_token_id)

    def skip_stopwords(self) -> "Tokens":
        return Tokens(list(filter(lambda token:token.tag != WordNet.STOPWORD.value,self.tokens)),pad_token_id=self.pad_token_id)
    
    def nouns(self) -> "Tokens":
        return Tokens(list(filter(lambda token:token.tag == WordNet.POS_NOUN.value, self.tokens)),pad_token_id=self.pad_token_id)

    def verbs(self) -> "Tokens":
        return Tokens(list(filter(lambda token:token.tag == WordNet.POS_VERB.value, self.tokens)),pad_token_id=self.pad_token_id)

    def entities(self) -> "Tokens":
        return Tokens(list(filter(lambda token:token.tag in (WordNet.POS_VERB.value,WordNet.POS_NOUN.value,WordNet.UNKNOWN.value,WordNet.SPECIAL.value), self.tokens)),pad_token_id=self.pad_token_id)

    def paraphrase(self) -> Generator[str, None, None]:
        for index,token in enumerate(self.tokens):
            for similar_token in token.similar_tokens:
                yield f"{' '.join(map(str,self.tokens[:index]))} {similar_token} {' '.join(map(str,self.tokens[index+1:]))}"
        yield str(self)

    def most_similar(self, others:List["Tokens"]) -> Optional["Tokens"]:
        if self.vector is None:
            return None
        others_with_vectors = filter(lambda other:other.vector is not None,others)
        others_vectors = array(list(map(
            lambda other:other.vector, 
            others_with_vectors
        )))
        index_best = argmax(others_vectors@self.vector)
        return others[index_best]

    def __sentence_embedding(self) -> ndarray:
        sparse_token_vectors = list(map(self.__embed_token,self.tokens))
        return concatenate([
            reduce(lambda vector1,vector2: vector1 & vector2, sparse_token_vectors),
            reduce(lambda vector1,vector2: vector1 | vector2, sparse_token_vectors),
            reduce(lambda vector1,vector2: vector1 ^ vector2, sparse_token_vectors),
        ])
        
    @staticmethod
    def __embed_token_phonetics(token:Token) -> ndarray:
        embedding = zeros(SIZE_METAPHONES,dtype=int)
        for character in token.phonology:
            metaphone_index = METAPHONES.index(character)
            embedding[metaphone_index] = 1
        return embedding

    @staticmethod
    def __embed_token_semantics(token:Token) -> ndarray:
        embedding = zeros(SIZE_SEMANTIC_VECTOR,dtype=int)
        if token.id >= SIZE_SEMANTIC_VECTOR:
            return embedding
        embedding[token.id] = 1
        for name in token.semantics:
            relation_index = VOCABULARY.index(name)
            embedding[relation_index] = 1
        return embedding
        
    @staticmethod
    def __embed_token(token:Token) -> ndarray:
        return concatenate([
            Tokens.__embed_token_phonetics(token),
            Tokens.__embed_token_semantics(token),
        ])    


from typing import Generator, Tuple
from enum import Enum 
from pathlib import Path 

import nltk

DATA_PATH = f"{Path(__file__).parent}/nltk_data/corpora"
if DATA_PATH not in nltk.data.path:
    nltk.data.path.append(DATA_PATH)

from nltk.corpus import stopwords, wordnet
from nltk.corpus.reader.wordnet import Synset
from nltk.corpus.reader.wordnet import ADJ, ADJ_SAT, ADV, NOUN, VERB

from ffast.preprocessor import PreprocessingPipeline

def create_vocabulary() -> Generator[Tuple[str,Synset],None,None]:
    for meaning in wordnet.all_synsets():
        for lemma in meaning.lemmas():
            yield (
                PREPROCESSOR.normalise(lemma.name()),
                meaning.name()
            )

PREPROCESSOR = PreprocessingPipeline()
STOPWORDS = sorted(stopwords.words('english'))
VOCABULARY_WORDNET = dict(create_vocabulary())
VOCABULARY = sorted(map(lambda meaning:meaning.name(),wordnet.all_synsets()))
SIZE_STOPWORDS = len(STOPWORDS)
SIZE_WORDNET = len(VOCABULARY)
METAPHONES = "ABCEFHIJKLMNOPRSTUWXY0. "
SIZE_METAPHONES = len(METAPHONES)
SIZE_SEMANTIC_VECTOR = SIZE_STOPWORDS + SIZE_WORDNET 
SIZE_WORD_VECTOR = SIZE_SEMANTIC_VECTOR + SIZE_METAPHONES 
SIZE_SENTENCE_VECTOR = 3*SIZE_WORD_VECTOR 

class WordNet(Enum):
    SKIP = "skip"
    UNKNOWN = "<Unknown>"
    SPECIAL = "<Special>"
    STOPWORD = "<StopWord>"
    POS_NOUN = "Noun"
    POS_VERB = "Verb"
    POS_MAP = {
        NOUN:"Noun",
        VERB:"Verb",
        ADV:"Adverb",
        ADJ:"Adjective",
        ADJ_SAT:"Adjective"
    }
    EXAMPLE_DELIMITER = "; "
    SYNSET_NAME_DELIMITER = "."


from typing import List, Generator

from nltk.util import ngrams

from ffast.wordnet.tokens import Token, Tokens
from ffast.wordnet.utils import (
    WordNet, PREPROCESSOR, 
    VOCABULARY, VOCABULARY_WORDNET, STOPWORDS,
    SIZE_STOPWORDS, SIZE_WORDNET, SIZE_SENTENCE_VECTOR
)
    
class Tokeniser:
    def __init__(self) -> None:
        self.size = SIZE_SENTENCE_VECTOR
        self.special_tokens = list()

    def __len__(self) -> int:
        return SIZE_WORDNET + SIZE_STOPWORDS + len(self.special_tokens) + 1

    def add_special_token(self, token:str) -> None:
        if token not in self.special_tokens and token not in VOCABULARY_WORDNET and token not in STOPWORDS:
            self.special_tokens.append(token)
            self.special_tokens.sort()

    def encode(self, text:str) -> Tokens:
        return Tokens(self._tokenise(text), pad_token_id=len(self)-1)

    def decode(self, ids:List[int]) -> Tokens:
        return Tokens(list(self._convert_ids_to_tokens(ids)),pad_token_id=len(self)-1)

    def _tokenise(self, text:str) -> List[Token]:
        words = text.split()
        number_of_words = len(words)
        tokens = [None]*number_of_words
        for ngram_size in range(number_of_words+1,0,-1):
            for index_start,ngram in enumerate(ngrams(words,ngram_size)):
                index_end = index_start + ngram_size
                if any(tokens[index_start:index_end]):
                    continue 
                raw_token = ' '.join(words[index_start:index_end])
                normalised_token = PREPROCESSOR.normalise(text=' '.join(ngram))
                synset_name = VOCABULARY_WORDNET.get(normalised_token)
                if synset_name is not None or ngram_size==1:
                    tokens[index_start:index_end] = [WordNet.SKIP.value]*ngram_size
                    tokens[index_start] = Token(
                        raw_token=raw_token,
                        normalised_token=normalised_token,
                        synset_name=synset_name,
                        sense_disambiguation_context=text,
                        SPECIAL_TOKENS=self.special_tokens
                    )
        return list(filter(lambda token:isinstance(token,Token),tokens))

    def _convert_ids_to_tokens(self, ids:List[int]) -> Generator[Token,None,None]:
        for id in ids:
            if id < SIZE_WORDNET:
                synset_name = VOCABULARY[id] 
                token = synset_name.split(WordNet.SYNSET_NAME_DELIMITER.value)[0]
            elif id < SIZE_WORDNET + SIZE_STOPWORDS:
                synset_name= None
                token = STOPWORDS[id-SIZE_WORDNET]
            elif id < SIZE_WORDNET + SIZE_STOPWORDS + len(self.special_tokens):
                synset_name= None
                id_special_token = id - (SIZE_WORDNET+SIZE_STOPWORDS)
                token = self.special_tokens[id_special_token] 
            else:
                synset_name= None
                token = WordNet.UNKNOWN.value                    
            yield Token(
                raw_token=token,
                normalised_token=PREPROCESSOR.preprocess(token),
                synset_name=synset_name,
                SPECIAL_TOKENS=self.special_tokens
            )


from setuptools import setup, find_packages
#from glob import glob 

VERSION = '0.4.6'
DESCRIPTION = 'FFAST: Fast Fourier Analysis for Sentence embeddings and Tokenisation'
LONG_DESCRIPTION = 'Fast and lightweight NLP pipeline for ML tasks: powerful tokeniser and (model-free) sentence embeddings using Fast Fourier transforms, power means, positional encoding and Wordnet or Poincare Embeddings'
#WORDNET_FILENAMES = glob('wordnet/nltk_data/corpora/wordnet/*')

setup(
    name="ffast",
    version=VERSION,
    author="Mohammed Terry-Jack",
    author_email="<mohammedterryjack@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    setup_requires = ['nltk'],
    include_package_data=True,
    package_data={'':[
        'poincare/poincare.txt',
        'wordnet/nltk_data/corpora/stopwords/english',
        'wordnet/nltk_data/corpora/stopwords.zip',
        'wordnet/nltk_data/corpora/wordnet.zip',
    ]},
    install_requires=['nltk', 'jellyfish', 'Unidecode', 'numpy', 'scipy'],
    keywords=['python', 'embedding', 'tokenisation', 'fast fourier', 'nlp', 'nlu', "poincare", "wordnet", "lite", "fast", "sentence encoder"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
    ]
)